# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 14:54:41 2024

@author: Jonathan
"""

from typing import Union
import sys
import re

import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets, QtGui

from .GUI_utilities import setLineEditBackground, MyLineEdit
from .icons import icons
from ..devices import DEVICES
from ..utilities import data_to_str, str_to_data, clean_string
from ..variables import (VARIABLES, get_variable, set_variable, Variable,
                         rename_variable, remove_variable, is_Variable,
                         has_variable, has_eval, eval_variable, EVAL)
from ..elements import Variable as Variable_og
from ..devices import get_element_by_address
from .GUI_instances import (openMonitor, openSlider, openPlotter,
                            closeMonitors, closeSliders, closePlotter,
                            clearVariablesMenu)


class VariablesMenu(QtWidgets.QMainWindow):

    variableSignal = QtCore.Signal(object)
    deviceSignal = QtCore.Signal(object)

    def __init__(self, has_parent: bool = False):

        super().__init__()
        self.has_parent = has_parent  # Only for closeEvent
        self.setWindowTitle('AUTOLAB - Variables Menu')
        self.setWindowIcon(icons['variables'])

        self.statusBar = self.statusBar()

        # Main widgets creation
        self.variablesWidget = QtWidgets.QTreeWidget(self)
        self.variablesWidget.setHeaderLabels(
            ['', 'Name', 'Value', 'Evaluated value', 'Type', 'Action'])
        self.variablesWidget.setAlternatingRowColors(True)
        self.variablesWidget.setIndentation(0)
        header = self.variablesWidget.header()
        header.setMinimumSectionSize(20)
        header.resizeSection(0, 20)
        header.resizeSection(1, 90)
        header.resizeSection(2, 120)
        header.resizeSection(3, 120)
        header.resizeSection(4, 50)
        header.resizeSection(5, 100)
        self.variablesWidget.itemDoubleClicked.connect(self.variableActivated)
        self.variablesWidget.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection)
        self.variablesWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.variablesWidget.customContextMenuRequested.connect(self.rightClick)

        addButton = QtWidgets.QPushButton('Add')
        addButton.clicked.connect(self.addVariableAction)

        removeButton = QtWidgets.QPushButton('Remove')
        removeButton.clicked.connect(self.removeVariableAction)

        self.devicesWidget = QtWidgets.QTreeWidget(self)
        self.devicesWidget.setHeaderLabels(['Name'])
        self.devicesWidget.setAlternatingRowColors(True)
        self.devicesWidget.setIndentation(10)
        self.devicesWidget.itemDoubleClicked.connect(self.deviceActivated)
        self.devicesWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.devicesWidget.customContextMenuRequested.connect(self.rightClickDevice)

        # Main layout creation
        layoutWindow = QtWidgets.QVBoxLayout()
        layoutWindow.setContentsMargins(0,0,0,0)
        layoutWindow.setSpacing(0)
        layoutTab = QtWidgets.QVBoxLayout()
        layoutWindow.addLayout(layoutTab)

        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(layoutWindow)
        self.setCentralWidget(centralWidget)

        refreshButtonWidget = QtWidgets.QPushButton()
        refreshButtonWidget.setText('Refresh Manager')
        refreshButtonWidget.clicked.connect(self.refresh)

        # Main layout definition
        layoutButton = QtWidgets.QHBoxLayout()
        layoutButton.addWidget(addButton)
        layoutButton.addWidget(removeButton)
        layoutButton.addStretch()

        frameVariables = QtWidgets.QFrame()
        layoutVariables = QtWidgets.QVBoxLayout(frameVariables)
        layoutVariables.addWidget(self.variablesWidget)
        layoutVariables.addLayout(layoutButton)

        frameDevices = QtWidgets.QFrame()
        layoutDevices = QtWidgets.QVBoxLayout(frameDevices)
        layoutDevices.addWidget(self.devicesWidget)

        tab = QtWidgets.QTabWidget(self)
        tab.addTab(frameVariables, 'Variables')
        tab.addTab(frameDevices, 'Devices')

        layoutTab.addWidget(tab)
        layoutTab.addWidget(refreshButtonWidget)

        self.resize(550, 300)
        self.refresh()

        # self.timer = QtCore.QTimer(self)
        # self.timer.setInterval(400) # ms
        # self.timer.timeout.connect(self.refresh_new)
        # self.timer.start()
        # VARIABLES.removeVarSignal.remove.connect(self.removeVarSignalChanged)
        # VARIABLES.addVarSignal.add.connect(self.addVarSignalChanged)

    def variableActivated(self, item: QtWidgets.QTreeWidgetItem):
        self.variableSignal.emit(item.name)

    def rightClick(self, position: QtCore.QPoint):
        """ Provides a menu where the user right clicked to manage a variable """
        item = self.variablesWidget.itemAt(position)
        if hasattr(item, 'menu'): item.menu(position)

    def rightClickDevice(self, position: QtCore.QPoint):
        """ Provides a menu where the user right clicked to manage a variable """
        item = self.devicesWidget.itemAt(position)
        if hasattr(item, 'menu'): item.menu(position)

    def deviceActivated(self, item: QtWidgets.QTreeWidgetItem):
        if hasattr(item, 'name'): self.deviceSignal.emit(item.name)

    def removeVariableAction(self):
        for variableItem in self.variablesWidget.selectedItems():
            remove_variable(variableItem.name)
            self.removeVariableItem(variableItem)

    # def addVariableItem(self, name):
    #     MyQTreeWidgetItem(self.variablesWidget, name, self)

    def removeVariableItem(self, item: QtWidgets.QTreeWidgetItem):
        index = self.variablesWidget.indexFromItem(item)
        self.variablesWidget.takeTopLevelItem(index.row())

    def addVariableAction(self):
        basename = 'var'
        name = basename
        names = list(VARIABLES)

        compt = 0
        while True:
            if name in names:
                compt += 1
                name = basename + str(compt)
            else:
                break

        variable = set_variable(name, 0)

        MyQTreeWidgetItem(self.variablesWidget, name, variable, self)  # not catched by VARIABLES signal

    # def addVarSignalChanged(self, key, value):
    #     print('got add signal', key, value)
    #     all_items = [self.variablesWidget.topLevelItem(i) for i in range(
    #         self.variablesWidget.topLevelItemCount())]

    #     for variableItem in all_items:
    #         if variableItem.name == key:
    #             variableItem.raw_value = get_variable(variableItem.name)
    #             variableItem.refresh_rawValue()
    #             variableItem.refresh_value()
    #             break
    #     else:
    #         self.addVariableItem(key)
    #     # self.refresh()  # TODO: check if item exists, create if not, update if yes

    # def removeVarSignalChanged(self, key):
    #     print('got remove signal', key)
    #     all_items = [self.variablesWidget.topLevelItem(i) for i in range(
    #         self.variablesWidget.topLevelItemCount())]

    #     for variableItem in all_items:
    #         if variableItem.name == key:
    #             self.removeVariableItem(variableItem)

    #     # self.refresh()  # TODO: check if exists, remove if yes

    def refresh(self):
        self.variablesWidget.clear()
        for var_name in VARIABLES:
            variable = get_variable(var_name)
            MyQTreeWidgetItem(self.variablesWidget, var_name, variable, self)

        self.devicesWidget.clear()
        for device_name, device in DEVICES.items():
            deviceItem = QtWidgets.QTreeWidgetItem(
                self.devicesWidget, [device_name])
            deviceItem.setBackground(0, QtGui.QColor('#9EB7F5'))  # blue
            deviceItem.setExpanded(True)
            for elements in device.get_structure():
                var = get_element_by_address(elements[0])
                MyQTreeWidgetItem(deviceItem, var.address(), var, self)

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modify the message displayed in the status bar and add error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)

    def closeEvent(self, event):
        # self.timer.stop()
        clearVariablesMenu()

        self.variablesWidget.clear()
        self.devicesWidget.clear()

        if not self.has_parent:
            closePlotter()
            closeMonitors()
            closeSliders()

            import pyqtgraph as pg
            try:
                # Prevent 'RuntimeError: wrapped C/C++ object of type ViewBox has been deleted' when reloading gui
                for view in pg.ViewBox.AllViews.copy().keys():
                    pg.ViewBox.forgetView(id(view), view)
                    # OPTIMIZE: forget only view used in monitor/gui
                pg.ViewBox.quit()
            except: pass

        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()

        super().closeEvent(event)

        if not self.has_parent:
            QtWidgets.QApplication.quit()  # close the app


class MyQTreeWidgetItem(QtWidgets.QTreeWidgetItem):

    def __init__(self,
                 itemParent: Union[QtWidgets.QTreeWidget,
                                   QtWidgets.QTreeWidgetItem],
                 name: str,
                 variable: Union[Variable, Variable_og],
                 gui: QtWidgets.QMainWindow):

        self.name = name
        self.variable = variable
        self.gui = gui

        if is_Variable(self.variable):
            super().__init__(itemParent, ['', name])
        else:
            super().__init__(itemParent, [name])
            return None

        nameWidget = QtWidgets.QLineEdit()
        nameWidget.setText(name)
        nameWidget.setAlignment(QtCore.Qt.AlignCenter)
        nameWidget.returnPressed.connect(self.renameVariable)
        nameWidget.textEdited.connect(lambda: setLineEditBackground(
            nameWidget, 'edited'))
        setLineEditBackground(nameWidget, 'synced')
        self.gui.variablesWidget.setItemWidget(self, 1, nameWidget)
        self.nameWidget = nameWidget

        rawValueWidget = MyLineEdit()
        rawValueWidget.setMaxLength(10000000)
        rawValueWidget.setAlignment(QtCore.Qt.AlignCenter)
        rawValueWidget.returnPressed.connect(self.changeRawValue)
        rawValueWidget.textEdited.connect(lambda: setLineEditBackground(
            rawValueWidget, 'edited'))
        self.gui.variablesWidget.setItemWidget(self, 2, rawValueWidget)
        self.rawValueWidget = rawValueWidget

        valueWidget = QtWidgets.QLineEdit()
        valueWidget.setMaxLength(10000000)
        valueWidget.setReadOnly(True)
        palette = valueWidget.palette()
        palette.setColor(QtGui.QPalette.Base,
                         palette.color(QtGui.QPalette.Base).darker(107))
        valueWidget.setPalette(palette)
        valueWidget.setAlignment(QtCore.Qt.AlignCenter)
        self.gui.variablesWidget.setItemWidget(self, 3, valueWidget)
        self.valueWidget = valueWidget

        typeWidget = QtWidgets.QLabel()
        typeWidget.setAlignment(QtCore.Qt.AlignCenter)
        self.gui.variablesWidget.setItemWidget(self, 4, typeWidget)
        self.typeWidget = typeWidget

        self.actionButtonWidget = None

        self.refresh_rawValue()
        self.refresh_value()

    def menu(self, position: QtCore.QPoint):
        """ This function provides the menu when the user right click on an item """
        menu = QtWidgets.QMenu()
        monitoringAction = menu.addAction("Start monitoring")
        monitoringAction.setIcon(icons['monitor'])
        monitoringAction.setEnabled(
            (hasattr(self.variable, 'readable')  # Action don't have readable
             and self.variable.readable
             and self.variable.type in (int, float, np.ndarray, pd.DataFrame)
             ) or (
                 is_Variable(self.variable)
                 and (has_eval(self.variable.raw) or isinstance(
                     self.variable.value, (int, float, np.ndarray, pd.DataFrame)))
                 ))

        plottingAction = menu.addAction("Capture to plotter")
        plottingAction.setIcon(icons['plotter'])
        plottingAction.setEnabled(monitoringAction.isEnabled())

        menu.addSeparator()
        sliderAction = menu.addAction("Create a slider")
        sliderAction.setIcon(icons['slider'])
        sliderAction.setEnabled(
            (hasattr(self.variable, 'writable')
             and self.variable.writable
             and self.variable.type in (int, float)))

        choice = menu.exec_(
            self.gui.variablesWidget.viewport().mapToGlobal(position))
        if choice == monitoringAction:
            openMonitor(self.variable, has_parent=True)
        if choice == plottingAction:
            openPlotter(variable=self.variable, has_parent=True)
        elif choice == sliderAction:
            openSlider(self.variable, gui=self.gui, item=self)

    def renameVariable(self) -> None:
        new_name = self.nameWidget.text()
        if new_name == self.name:
            setLineEditBackground(self.nameWidget, 'synced')
            return None

        if new_name in VARIABLES:
            self.gui.setStatus(
                f"Error: {new_name} already exist!", 10000, False)
            return None

        new_name = clean_string(new_name)

        try:
            rename_variable(self.name, new_name)
        except Exception as e:
            self.gui.setStatus(f'Error: {e}', 10000, False)
        else:
            self.name = new_name
            new_name = self.nameWidget.setText(self.name)
            setLineEditBackground(self.nameWidget, 'synced')
            self.gui.setStatus('')
        return None

    def refresh_rawValue(self):
        raw_value_str = data_to_str(self.variable.raw)

        self.rawValueWidget.setText(raw_value_str)
        setLineEditBackground(self.rawValueWidget, 'synced')

        if has_variable(self.variable.raw):  # OPTIMIZE: use hide and show instead but doesn't hide on instantiation
            if self.actionButtonWidget is None:
                actionButtonWidget = QtWidgets.QPushButton()
                actionButtonWidget.setText('Update value')
                actionButtonWidget.setMinimumSize(0, 23)
                actionButtonWidget.setMaximumSize(85, 23)
                actionButtonWidget.clicked.connect(self.convertVariableClicked)
                self.gui.variablesWidget.setItemWidget(self, 5, actionButtonWidget)
                self.actionButtonWidget = actionButtonWidget
        else:
            self.gui.variablesWidget.removeItemWidget(self, 5)
            self.actionButtonWidget = None

    def refresh_value(self):
        value = self.variable.value
        value_str = data_to_str(value)

        self.valueWidget.setText(value_str)
        self.typeWidget.setText(str(type(value)).split("'")[1])


    def changeRawValue(self):
        name = self.name
        raw_value = self.rawValueWidget.text()
        try:
            if not has_eval(raw_value):
                raw_value = str_to_data(raw_value)
            else:
                # get all variables
                raw_value_check = raw_value[len(EVAL): ]  # Allows variable with name 'eval'
                pattern1 = r'[a-zA-Z][a-zA-Z0-9._]*'
                matches1 = re.findall(pattern1, raw_value_check)
                # get variables not unclosed by ' or " (gives bad name so needs to check with all variables)
                pattern2 = r'(?<!["\'])([a-zA-Z][a-zA-Z0-9._]*)(?!["\'])'
                matches2 = re.findall(pattern2, raw_value_check)
                matches = list(set(matches1) & set(matches2))
                # Add device/variable name to matches
                for match in list(matches):
                    matches.append(match.split('.')[0])
                matches = list(set(matches))
                assert name not in matches, f"Variable '{name}' name can't be used in eval to avoid circular definition"
        except Exception as e:
            self.gui.setStatus(f'Error: {e}', 10000, False)
        else:
            try: set_variable(name, raw_value)
            except Exception as e:
                self.gui.setStatus(f'Error: {e}', 10000)
            else:
                self.refresh_rawValue()
                self.refresh_value()

    def convertVariableClicked(self):
        try: value = eval_variable(self.variable)
        except Exception as e:
            self.gui.setStatus(f'Error: {e}', 10000, False)
        else:
            value_str = data_to_str(value)

            self.valueWidget.setText(value_str)
            self.typeWidget.setText(str(type(value)).split("'")[1])
            # self.gui.refresh()  # OPTIMIZE replace by each variable send update signal



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:15:26 2019

@author: qchat
"""

import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets, QtGui

from .customWidgets import MyQTreeWidget, MyQTabWidget, MyQTreeWidgetItem
from ..icons import icons
from ..GUI_utilities import MyInputDialog, qt_object_exists
from ...config import get_scanner_config
from ...variables import has_eval
from ...utilities import (clean_string, str_to_array, array_to_str,
                          str_to_dataframe, dataframe_to_str, str_to_tuple)


class RecipeManager:
    """ Manage a recipe from a scan """

    def __init__(self, gui: QtWidgets.QMainWindow, recipe_name: str):

        self.gui = gui  # gui is scanner
        self.recipe_name = recipe_name

        # Import Autolab config
        scanner_config = get_scanner_config()
        self.precision = scanner_config['precision']

        self.defaultItemBackground = None

        # Recipe frame
        frameRecipe = QtWidgets.QFrame()

        # Tree configuration
        self.tree = MyQTreeWidget(frameRecipe, self.gui, self.recipe_name)
        self.tree.setHeaderLabels(['Step name', 'Action', 'Element address',
                                   'Type', 'Value', 'Unit'])
        header = self.tree.header()
        header.setMinimumSectionSize(20)
        header.setStretchLastSection(False)
        header.resizeSection(0, 95)
        header.resizeSection(1, 55)
        header.resizeSection(2, 115)
        header.resizeSection(3, 35)
        header.resizeSection(4, 110)
        header.resizeSection(5, 32)
        header.setMaximumSize(16777215, 16777215)
        self.tree.itemDoubleClicked.connect(self.itemDoubleClicked)
        self.tree.reorderSignal.connect(self.orderChanged)
        self.tree.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
        self.tree.setDropIndicatorShown(True)
        self.tree.setAlternatingRowColors(True)
        self.tree.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tree.setSizePolicy(
            QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Ignored)
        self.tree.customContextMenuRequested.connect(self.rightClick)
        self.tree.setMinimumSize(0, 200)
        self.tree.setMaximumSize(16777215, 16777215)
        self.tree.setIndentation(0)

        layoutRecipe = QtWidgets.QVBoxLayout(frameRecipe)
        layoutRecipe.addWidget(self.tree)

        # Qframe and QTab for close+parameter+scanrange+recipe
        frameAll = QtWidgets.QFrame()
        layoutAll = QtWidgets.QVBoxLayout(frameAll)
        layoutAll.setContentsMargins(0,0,0,0)
        layoutAll.setSpacing(0)

        layoutAll.addWidget(frameRecipe, stretch=1)

        frameAll2 = MyQTabWidget(frameAll, self.gui, self.recipe_name)
        self._frame = frameAll2
        self._layoutAll = layoutAll
        self.gui.verticalLayout_recipe.addWidget(self._frame)

    def __del__(self):
        self._removeWidget()

    def _removeWidget(self):
        if hasattr(self, '_frame'):
            try:
                self._frame.hide()
                self._frame.deleteLater()
                del self._frame
            except: pass

    def _activateTree(self, active: bool):
        self._frame.setTabEnabled(0, bool(active))

    def orderChanged(self, event):
        newOrder = [self.tree.topLevelItem(i).text(0)
                    for i in range(self.tree.topLevelItemCount())]
        self.gui.configManager.setRecipeStepOrder(self.recipe_name, newOrder)

    def refresh(self):
        """ Refreshs the whole scan recipe displayed from the configuration center """
        self.tree.clear()

        for step in self.gui.configManager.stepList(self.recipe_name):
            # Loading step informations
            item = MyQTreeWidgetItem(self.tree, step, self)
            self.defaultItemBackground = item.background(0)

        self.tree.setFocus()  # needed for ctrl+z ctrl+y on recipe
        # toggle recipe
        active = bool(self.gui.configManager.getActive(self.recipe_name))
        self.gui._activateRecipe(self.recipe_name, active)

    def rightClick(self, position: QtCore.QPoint):
        """ Provides a menu where the user right clicked to manage a recipe """
        if not self.gui.scanManager.isStarted():
            item = self.tree.itemAt(position)

            if item is not None:
                name = item.text(0)
                stepType = self.gui.configManager.getRecipeStepType(
                    self.recipe_name, name)
                element = self.gui.configManager.getRecipeStepElement(
                    self.recipe_name, name)

                menuActions = {}
                menu = QtWidgets.QMenu()
                menuActions['copy'] = menu.addAction("Copy")
                menuActions['copy'].setIcon(icons['copy'])
                menuActions['copy'].setShortcut(QtGui.QKeySequence("Ctrl+C"))
                menu.addSeparator()

                if stepType == 'set' or (stepType == 'action' and element.type in [
                        int, float, bool, str, tuple, np.ndarray, pd.DataFrame]):
                    menuActions['setvalue'] = menu.addAction("Set value")
                    menuActions['setvalue'].setIcon(icons['write'])

                menuActions['remove'] = menu.addAction("Remove")
                menuActions['remove'].setIcon(icons['remove'])
                menuActions['remove'].setShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Delete))

                menuActions['rename'] = menu.addAction("Rename")
                menuActions['rename'].setIcon(icons['rename'])
                menuActions['rename'].setShortcut(QtGui.QKeySequence("Ctrl+R"))

                choice = menu.exec_(self.tree.viewport().mapToGlobal(position))

                if choice == menuActions['copy']:
                    self.copyStep(name)
                elif choice == menuActions['rename']:
                    self.renameStep(name)
                elif choice == menuActions['remove']:
                    self.removeStep(name)
                elif 'setvalue' in menuActions and choice == menuActions['setvalue']:
                    self.setStepValue(name)
            else:
                menuActions = {}
                menu = QtWidgets.QMenu()
                menuActions['paste'] = menu.addAction("Paste")
                menuActions['paste'].setIcon(icons['paste'])
                menuActions['paste'].setShortcut(QtGui.QKeySequence("Ctrl+V"))
                menuActions['paste'].setEnabled(self.gui._copy_step_info is not None)
                choice = menu.exec_(self.tree.viewport().mapToGlobal(position))

                if choice == menuActions['paste']:
                    self.pasteStep()

            # TODO: disabled this feature has it is not good in its current state
            #     config = self.gui.configManager.config
            #     if len(config) > 1:

            #         recipe_name_list = self.gui.configManager.getAllowedRecipe(self.recipe_name)

            #         menuActions = {}
            #         menu = QtWidgets.QMenu()
            #         for recipe_name in recipe_name_list:
            #             menuActions[recipe_name] = menu.addAction(f'Add {recipe_name}')
            #             menuActions[recipe_name].setIcon(icons['recipe'])

            #         if len(recipe_name_list) != 0:
            #             choice = menu.exec_(self.tree.viewport().mapToGlobal(position))

            #         for recipe_name in recipe_name_list:
            #             if choice == menuActions[recipe_name]:
            #                 self.gui.configManager.configHistory.active = False
            #                 active = self.gui.configManager.getActive(recipe_name)

            #                 if active: self.gui.configManager.activateRecipe(
            #                         recipe_name, not active)  # disable

            #                 for param in self.gui.configManager.parameterList(recipe_name):
            #                     newName = self.gui.configManager.getUniqueName(self.recipe_name, param['name'])
            #                     self.gui.configManager.renameParameter(
            #                         recipe_name, param['name'], newName)

            #                 for step in self.gui.configManager.stepList(recipe_name):
            #                     newName = self.gui.configManager.getUniqueName(self.recipe_name, step['name'])
            #                     self.gui.configManager.renameRecipeStep(
            #                         recipe_name, step['name'], newName)

            #                 self.gui.configManager.configHistory.active = True

            #                 self.gui.configManager.addRecipeStep(
            #                     self.recipe_name, 'recipe',
            #                     recipe_name,  f'do_{recipe_name}')

            #                 break

    def copyStep(self, name: str):
        """ Copy step information """
        self.gui._copy_step_info = self.gui.configManager.getRecipeStep(
            self.recipe_name, name)

    def pasteStep(self):
        """ Paste step information """
        step_info = self.gui._copy_step_info
        if step_info is not None:
            self.gui.configManager.addRecipeStep(
                self.recipe_name, step_info['stepType'],
                step_info['element'], step_info['name'],
                value=step_info['value'])

    def renameStep(self, name: str):
        """ Prompts the user for a new step name and apply it to the selected step """
        newName, state = QtWidgets.QInputDialog.getText(
            self.gui, name, f"Set {name} new name",
            QtWidgets.QLineEdit.Normal, name)

        newName = clean_string(newName)
        if newName != '':
            self.gui.configManager.renameRecipeStep(
                self.recipe_name, name, newName)

    def removeStep(self, name: str):
        self.gui.configManager.delRecipeStep(self.recipe_name, name)

    def setStepValue(self, name: str):
        """ Prompts the user for a new step value and apply it to the selected step """
        element = self.gui.configManager.getRecipeStepElement(
            self.recipe_name, name)
        value = self.gui.configManager.getRecipeStepValue(
            self.recipe_name, name)
        # Default value displayed in the QInputDialog
        if has_eval(value):
            defaultValue = f'{value}'
        else:
            if element.type in [np.ndarray]:
                defaultValue = array_to_str(value, threshold=1000000, max_line_width=100)
            elif element.type in [pd.DataFrame]:
                defaultValue = dataframe_to_str(value, threshold=1000000)
            if element.type in [bytes] and isinstance(value, bytes):
                defaultValue = f'{value.decode()}'
            else:
                try:
                    defaultValue = f'{value:.{self.precision}g}'
                except (ValueError, TypeError):
                    defaultValue = f'{value}'

        main_dialog = MyInputDialog(self.gui, name)
        main_dialog.setWindowModality(QtCore.Qt.ApplicationModal)  # block GUI interaction
        main_dialog.setTextValue(defaultValue)
        main_dialog.show()

        if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
            value = main_dialog.textValue()

            try:
                try:
                    assert has_eval(value), "Need $eval: to evaluate the given string"
                except:
                    # Type conversions
                    if element.type in [int]:
                        value = int(value)
                    elif element.type in [float]:
                        value = float(value)
                    elif element.type in [str]:
                        value = str(value)
                    elif element.type in [bytes]:
                        value = value.encode()
                    elif element.type in [bool]:
                        if value == "False": value = False
                        elif value == "True": value = True
                        value = int(value)
                        assert value in [0, 1]
                        value = bool(value)
                    elif element.type in [tuple]:
                        value = str_to_tuple(value)
                    # OPTIMIZE: bad with large data (truncate), but nobody will use it for large data right?
                    elif element.type in [np.ndarray]:
                        value = str_to_array(value)
                    elif element.type in [pd.DataFrame]:
                        value = str_to_dataframe(value)
                    else:
                        assert has_eval(value), "Need $eval: to evaluate the given string"
                # Apply modification
                self.gui.configManager.setRecipeStepValue(
                    self.recipe_name, name, value)
            except Exception as er:
                self.gui.setStatus(f"Can't set step: {er}", 10000, False)
                pass

        main_dialog.deleteLater()

    def itemDoubleClicked(self, item: QtWidgets.QTreeWidgetItem, column: int):
        """ Executes an action depending where the user double clicked """
        if not self.gui.scanManager.isStarted():
            name = item.text(0)
            stepType = self.gui.configManager.getRecipeStepType(self.recipe_name, name)
            element = self.gui.configManager.getRecipeStepElement(self.recipe_name, name)

            if column == 0:
                self.renameStep(name)
            elif column == 4:
                if stepType == 'set' or (stepType == 'action' and element.type in [
                        int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]):
                    self.setStepValue(name)

    def setStepProcessingState(self, name: str, state: str):
        """ Sets the background color of a recipe step during the scan """
        if not qt_object_exists(self.tree):
            return None
        item = self.tree.findItems(name, QtCore.Qt.MatchExactly, 0)[0]

        if state is None:
            item.setBackground(0, self.defaultItemBackground)
        elif state == 'started':
            item.setBackground(0, QtGui.QColor('#ff8c1a'))
        elif state == 'finished':
            item.setBackground(0, QtGui.QColor('#70db70'))

    def resetStepsProcessingState(self):
        """ Resets the background color of a recipe once the scan is finished """
        for name in self.gui.configManager.getNames(self.recipe_name, option='recipe'):
            self.setStepProcessingState(name, None)


#
#class Step:
#
#    def __init__(self,name):
#        self.name = None
#
#    def set_name(self,name):
#        self.name = name
#
#    def message(self,mess):
#        return f'Step {self.name} (self.step_type): {mess}'
#
#
#class ExecuteActionStep(Step):
#    step_type = 'execute_action'
#
#    def __init__(self,name,element,value=None):
#        Step.__init__(self,name)
#
#        # Element
#        assert element._element_type == 'action', self.message('The element has to be an Action')
#        self.action = element
#
#        # Value
#        if self.action.has_parameter :
#            assert value is not None, self.message('Parameter value required')
#            self.set_value(value)
#        else :
#            assert value is None, self.message('This action has no parameter')
#
#    def set_value(self,value):
#        assert self.action.has_parameter, self.message('This action has no parameter')
#        try :
#            self.value = self.action.type(value)
#        except :
#            raise ValueError(self.message(f'Impossible to convert {value} in type {self.element.type.__name__}'))
#
#    def get_value(self):
#        return self.value
#
#    def run(self):
#        if self.element.has_parameter :
#            assert self.value is not None
#            self.action(self.value)
#        else :
#            self.action()
#
#
#
#class ScanVariableStep:
#    step_type = 'scan_variable'
#
#    def __init__(self,name,element):
#
#        Step.__init__(self,name)
#
#        # Element
#        assert element._element_type == 'variable', self.message('The element has to be a Variable')
#        assert element.writable is True, self.message('The Variable has to be writable')
#        self.variable = element
#
#
#class ScanVariableEndStep:
#    step_type = 'scan_variable_end'
#
#    def __init__(self,name):
#
#
#    def run(self):
#
#
#
#class MeasureVariableStep:
#    step_type = 'measure_variable'
#
#    def __init__(self,name,element):
#        Step.__init__(self,name)
#
#        # Element
#        assert element._element_type == 'variable', self.message('The element has to be a Variable')
#        assert element.readable is True, self.message('The Variable has to be readable')
#        self.variable = element
#
#
#    def run(self):
#        return self.variable()
#
#
#
#
#class SetVariableStep:
#    step_type = 'set_variable'
#
#    def __init__(self,name,element,value):
#        Step.__init__(self,name)
#
#        # Element
#        assert element._element_type == 'variable', self.message('The element has to be a Variable')
#        assert element.writable is True, self.message('The Variable has to be writable')
#        self.variable = element
#
#        # Value
#        self.set_value(value)
#
#
#    def set_value(self,value):
#        try :
#            self.value = self.variable.type(value)
#        except :
#            raise ValueError(self.message(f'Impossible to convert {value} in type {self.element.type.__name__}'))
#
#    def get_value(self):
#        return self.value
#
#    def run(self):
#        self.variable(self.value)
#
#
#    def prompt_value(self):
#
#        # Default value displayed in the QInputDialog
#        if self.variable.type == str :
#            defaultValue = f'{self.value}'
#        else :
#            defaultValue = f'{self.value:g}'
#
#        value,state = QtWidgets.QInputDialog.getText(self.gui,
#                                                     name,
#                                                     f"Set {name} value",
#                                                     QtWidgets.QLineEdit.Normal, defaultValue)
#
#        if value != '' :
#
#            try :
#
#                # Type conversions
#                if element.type in [int]:
#                    value = int(value)
#                elif element.type in [float] :
#                    value = float(value)
#                elif element.type in [str] :
#                    value = str(value)
#                elif element.type in [bool]:
#                    value = int(value)
#                    assert value in [0,1]
#                    value = bool(value)
#
#                # Apply modification
#                self.gui.configManager.setRecipeStepValue(name,value)
#
#            except :
#                pass
#
